/* sp.pl */
/* Construct specialised program */
/* 
	1.  Flow analysis performed by abint/3
	2.  Renaming of predicates performed by rename_preds/4
	3.  Construction of new procedures by make_clauses/5
	4.  Writing out to file by write_to_file/1.

*/
/* This core procedure can be adapted to perform
   various kinds of specialisation.  Different abstraction
   operations, and unfolding transformations can be defined.
   The following procedures define variants of the program:

	resultant/2
	more_specific/1
	cleanup/3

 */

/* John Gallagher: 31.7.90 */
/*  revised 31.8.90 */

%:- setdebug.

%:- dynamic directive/1.

sp(G) :-
	flow_analyse(G,Tree,_),
	rename_preds(G,Tree,R),
	!,
	make_clauses(G,R,Cls),
	!,
	write_to_file(Cls),
        !.


flow_analyse(G,Tree,Pathtable) :-
        nl,
        write('flow analysis---'),sp_ttyflush,
        abint(G,Tree,Pathtable).

rename_preds(G,Tree,R) :-
	treeroot(Tree,Initrep),
	functor(G,P,_),
	R = [name(Initrep,0,P)|_],
	renames(Tree,R).

treeroot([Branch|_],G) :-
	root(Branch,G).

root([(G,_)],G).
root([_|Br],G) :-
	root(Br,G).

make_clauses(G,R,Cls) :-
        nl,write('constructing specialised program---'),sp_ttyflush,
        procedures(R,R,Cls1),
        nl,nl,write('cleaning up---'),sp_ttyflush,
        cleanup(G,Cls1,Cls).

write_to_file(Cls) :-
        nl,write('Enter filename: '),
        read(F),
        sp_filewrite(F,Cls).

/* Rename each atom in the environment */

renames([],R) :-
	list1(R).
renames([[]|Tree],R) :-
	renames(Tree,R).
renames([[loop|Br]|Tree],R) :-
	!,
	renames([Br|Tree],R).
renames([[fail|Br]|Tree],R) :-
	!,
	renames([Br|Tree],R).
renames([[(B,_)|Br]|Tree],R) :-
	renamepred(B,R),
	renames([Br|Tree],R).
	
renamepred(P,S) :-
	renamepred(P,S,0).

renamepred(A,S,M) :-
	var(S),
	!,
	N is M + 1,
	A =.. [P|_],
	name(P,Plist),
	name(N,Nlist),
	my_append(Plist,[95|Nlist],PNlist), /* 95 = underscore */
	name(Pn,PNlist),
	S = [name(A,N,Pn)|_],
	!.
renamepred(P,[name(P,_,_)|_],_) :-
	!.
renamepred(P,[name(Q,M,_)|S],_) :-
	P =.. [A|_],
	Q =.. [A|_],
	!,
	renamepred(P,S,M).
renamepred(P,[_|S],K) :-
	renamepred(P,S,K).

/*  Generate procedures for the calls  */

procedures([],_,[]) :-
	!.
procedures([name(A,N,Pn)|Rs],R,[Cls|Prs]) :-
        procedure(A,R,N,Pn,Cls),
	!,
        procedures(Rs,R,Prs).

procedure(A,Rs,0,_,Cls) :- /* no renaming of initial goals */
	!,
	nl,nl,write(A),
        melt(A,Ma),
        bagof1((Ma :- R), resultant(Ma,R), Cls1),
	rename_bodies(Cls1,Rs,Cls).
procedure(A,Rs,_,Pn,Cls) :-
	nl,write(A),
        melt(A,Ma),
        vars(Ma,Vs),
	sp_length(Vs,Nargs),
	Nargs < 21,
	!,
	%nonvarargs(Ma,Vs),
	%Ma =.. [_|Xs],
	%level1_args(Xs,Ws,[]),
	%nonvarargs1(Ws,Vs),
	H =.. [Pn|Vs],
        bagof1((H :- R), Ma^resultant(Ma,R), Cls1),
	rename_bodies(Cls1,Rs,Cls).

procedure(A,Rs,_,Pn,Cls) :-
	nl,write(A),
        melt(A,Ma),
	Ma =.. [_|Xs],
	H =.. [Pn|Xs],
        bagof1((H :- R), Ma^resultant(Ma,R), Cls1),
	rename_bodies(Cls1,Rs,Cls).

nonvarargs(M,Vs) :-
	M =.. [_|Xs],
	nonvarargs1(Xs,Vs).

nonvarargs1([],[]).
nonvarargs1([X|Xs],Vs) :-
	ground(X),
	!,
	nonvarargs1(Xs,Vs).
nonvarargs1([X|Xs],[X|Vs]) :-
	nonvarargs1(Xs,Vs).

level1_args([],L,L).
level1_args([X|Xs],L,L2) :-
	nonvar(X),
	!,
	X =.. [_|Ys],
	d_list(Ys,L,L1),
	level1_args(Xs,L1,L2).
level1_args([X|Xs],[X|L],L1) :-
	level1_args(Xs,L,L1).

d_list([],L,L).
d_list([X|Xs],[X|L],L1) :-
	d_list(Xs,L,L1).

rename_bodies([],_,[]) :-
	!.
rename_bodies([(H :- R1)|Cls1],Rs,[(H :- R2)|Cls]) :-
        make_body(R1,Rs,R2),
	!,
	rename_bodies(Cls1,Rs,Cls).
rename_bodies([_|Cls1],Rs,Cls) :-  	/* couldn't make body */
	rename_bodies(Cls1,Rs,Cls).	/* because it fails   */

make_body(B,_,B) :- 
	var(B),
	!.
make_body((B,Bs),Rs,(A,As)) :- 
	!,
        make_body(B,Rs,A),
        make_body(Bs,Rs,As).
make_body(B,Rs,B1) :- 
	prolog_meta(B,B1,P,P1),
	!,
	make_body(P,Rs,P1).
make_body(B,Rs,A) :-
        buildNewGoal(B,Rs,A).

buildNewGoal(G,_,G) :-
        sp_builtin(G),
        !.
buildNewGoal(G,_,G) :-
        imported(G),
        !.
buildNewGoal(B,Rs,A) :-
        copyterm(B,Cb),
	canonical(Cb),
        rep(Cb,R,Rs),
        renameGoal(R,B,A,Rs).

renameGoal(fail,_,fail,_) :-
	!.
renameGoal(R,A,A,Rs) :-
        memb2(name(R,0,_),Rs),
	!.
renameGoal(R,B,A,Rs) :-
        memb2(name(R,_,Pn),Rs),
	melt(R,Mr),
	vars(Mr,Vs),
	sp_length(Vs,Nargs),
	Nargs < 21,
	!,
	%nonvarargs(Mr,Vs),
	%Mr =.. [_|Xs],
	%level1_args(Xs,Ws,[]),
	%nonvarargs1(Ws,Vs),
	A =.. [Pn|Vs],
	B = Mr,
	!.
renameGoal(R,B,A,Rs) :-
        memb2(name(R,_,Pn),Rs),
	melt(R,Mr),
	Mr =.. [_|Xs],
	A =.. [Pn|Xs],
	B = Mr,
	!.
renameGoal(_,B,_,_) :-
	nl,
        write(B),
        write('cannot be renamed.......'),
        nl,nl,
        fail.

 
rep(A,B,Rs) :-
	find_instance(A,C,Rs),
	!,
	most_specific_rep(A,C,B,Rs).
rep(_,fail,_) :-
	nl,
	write('Failed atom detected:  replaced by ''fail'''),
	nl.

find_instance(A,B,[name(B,_,_)|_]) :-
	instance_of(A,B),
	!.
find_instance(A,C,[_|Bs]) :-
	find_instance(A,C,Bs).

most_specific_rep(D,A,C,[name(B,_,_)|Bs]) :-
	instance_of(B,A),
	instance_of(D,B),
	most_specific_rep(D,B,C,Bs).
most_specific_rep(D,A,C,[_|Bs]) :-
	most_specific_rep(D,A,C,Bs).
most_specific_rep(_,A,A,[]).

sp_filewrite(tty,Cls) :-
        !,
	nl,write('Specialised Program:'),nl,nl,
        displayqProcs(Cls),
	nl,nl.
sp_filewrite(F,Cls) :-
        tell(F),
	write_directives,
        displayqProcs(Cls),
        told.

/* write the program to current output file  */

displayqProcs([P|Ps]) :-
        displayqClauses(P),
        displayqProcs(Ps).
displayqProcs([]).

write_directives :-
	bagof(X,(directive(X),X \== dummy,write_directive(X)),_),
	!.
write_directives.

write_directive(X) :-
	canonical(X),
	write(':- '),
	write(X),
	write('.'),
	nl.
