/* Compute the characteristic path of a call:
  This version computes a long path, according to the following
  algorithm.

	repeat
		determinate path; msg of choices
	until resolvent is true

  A length bound on the path can be imposed. */

/* JPG: 10.7.90 */
/* based on callpath.pl */

/****************************************************/
/* export:  charpath/3,                             */
/* export:  resultant/2,                            */
/* export:  more_specific/1,                        */
/* export:  cleanup/3,                              */
/****************************************************/


charpath(G,Path) :-
	repeat_callpath(G,Path).

repeat_callpath(true,[]) :-
	!.
repeat_callpath(Gs,[Path,Choice|Ps]) :-
	callpath(Gs,Path,Choice,R1),
	solve_choice(R1,R2),
	!,
	repeat_callpath(R2,Ps).
	
callpath(G,Cs,Choice,R) :-
	prefix(G,Cs,[],R),
	choice(R,Choice).

resultant(A,R) :-
	clause(A,R1),
	prefix(R1,_,_,R2),
	unfold_choice(R2,R),
	more_specific(R).


more_specific(B) :-
	var(B),
	!.
more_specific((B,Bs)) :-
	!,
	more_specific(B),
	more_specific(Bs).
more_specific(B) :-
        sp_builtin(B),
        unfoldable(B),
        !,
        B.
more_specific(B) :-
        sp_builtin(B),
        !.
more_specific(B) :-
        imported(B),
        !.
more_specific(B) :-
        callpath(B,_,_,R),
        solve_choice(R,_).

lub(A,B,C) :-
	melt(A,A1),
	melt(B,B1),
	env_lub(A1,B1,C),
	canonical(C).

env_lub(env(A,G,F),env(A1,G1,F1),env(A2,G2,F2)) :-
	msg1(A,A1,A2,VMap),
	new_env(VMap,G,F,G1,F1,G2,[],F2,[]).

new_env([],_,_,_,_,G,G,F,F) :-
	!.
new_env([subst(T1,T2,V)|Vs],G1,F1,G2,F2,G,G3,F,F3) :-
	mode_classify(T1,G1,F1,M1),
	mode_classify(T2,G2,F2,M2),
	mode_lub(M1,M2,M3),
	add_env(M3,V,G,G4,F,F4),
	new_env(Vs,G1,F1,G2,F2,G4,G3,F4,F3).

add_env(unknown,_,G,G,F,F).
add_env(ground,V,[V|G],G,F,F).
add_env(free,V,G,G,[V|F],F).
	

mode_lub(ground,ground,ground) :-
	!.
mode_lub(free,free,free) :-
	!.
mode_lub(_,_,unknown).

mode_classify(T,G,_,ground) :-
	ground_mode(T,G),
	!.
mode_classify(T,_,F,free) :-
	free_mode(T,F),
	!.
mode_classify(_,_,_,unknown).

ground_mode(X,G) :-
	var(X),
	!,
	varmemb(X,G).
ground_mode(T,G) :-
	T =.. [_|Xs],
	ground_mode_list(Xs,G).

ground_mode_list([],_) :-
	!.
ground_mode_list([X|Xs],G) :-
	ground_mode(X,G),
	ground_mode_list(Xs,G).

free_mode(X,F) :-
	var(X),
	!,
	varmemb(X,F).

varmemb(X,[Y|_]) :-
	X == Y,
	!.
varmemb(X,[_|Ys]) :-
	varmemb(X,Ys).



prefix(X,[var|Cs],Cs,X) :-
	var(X),
	!.
prefix(true,Cs,Cs,true) :-
	!.
prefix((G,Gs),Ns,Ns2,Gs2) :-
	!,
	prefix(G,Ns,Ns1,Gs1),
	prefix1(Gs1,Gs,Ns1,Ns2,Gs2).
prefix(G,Ns,Ns,G) :-
	imported(G),
	!.
prefix(clause(H,B),Ms,Ns,G1) :-
	!,
	choice1(clause(H,B),Chs),
	det_check(Chs,clause(H,B),Ms,Ns,G1).
prefix(G,Ms,Ns,G1) :-
	sp_builtin(G),
	!,
	prefix2(G,Ms,Ns,G1).
prefix(G,Ms,Ns2,Gs1) :-
	choice1(G,Chs),
	det_check(Chs,G,Ms,Ns2,Gs1).

prefix1(true,Gs,Ns,Ns1,G1) :-
	!,
	prefix(Gs,Ns,Ns1,G1).
prefix1(G,Gs,Ns,Ns,Gs1) :-
	joingoals(G,Gs,Gs1).

prefix2(G,[P|Ns],Ns,true) :-
	unfoldable(G),
	!,
	choice1(G,P),
	call(G).
prefix2(G,Ns,Ns,G).
	
det_check([N],clause(H,B),[N|Ns],Ns,true) :-
	!,
	clause(H,B).
det_check([N],G,[N|Ns],Ns2,Gs1) :-
	!,
	clause(G,B),
	prefix(B,Ns,Ns2,Gs1).
det_check([_,_|_],G,Ns,Ns,G).

choice((G,_),Cs) :-
	!,
	choice1(G,Cs).
choice(G,Cs) :-
	choice1(G,Cs).

choice1(clause(G,B),Cs) :-
	!,
	bagof(N,[G,B]^(ref_clause(G,B,N),no_Deadend(B)),Cs).
choice1(call(G),P) :-
	!,
	pattern(G,P).
choice1(G,P) :-
	imported(G),
	!,
	functor(G,P,_).
choice1(G,P) :-
	sp_builtin(G),
	!,
	pattern(G,P).
choice1(G,Cs) :-
	bagof(N,[G,B]^(ref_clause(G,B,N),no_Deadend(B)),Cs).

?- re_read([cleanup]).
/*
?- [cleanup].
*/
