/* sp.pl */
/* Construct specialised program */
/* 
	1.  Flow analysis performed by abint/3
	2.  Renaming of predicates performed by rename_preds/4
	3.  Construction of new procedures by make_clauses/5
	4.  Writing out to file by write_to_file/1.

*/
/* This core procedure can be adapted to perform
   various kinds of specialisation.  Different abstraction
   operations, and unfolding transformations can be defined.
   The following procedures define variants of the program:

	resultant/2
	more_specific/1
	cleanup/3

 */

/* John Gallagher: 31.7.90 */
/*  revised 31.8.90 */


sp(G) :-
	flow_analyse(G,Pfg,Pathtable),
	rename_preds(G,Pfg,Pathtable,R),
	!,
	make_clauses(G,R,Pfg,Pathtable,Cls),
	!,
	write_to_file(Cls),
        !.


flow_analyse(G,Pfg,Pathtable) :-
        nl,
        write('flow analysis---'),sp_ttyflush,
        abint(G,Pfg,Pathtable).

rename_preds(G,Pfg,Pathtable,R) :-
	rep(G,Initrep,Pathtable,Pfg),
	functor(G,P,_),
	R = [name(Initrep,0,P)|_],
	renames(Pfg,R).

make_clauses(G,R,Pfg,Pathtable,Cls) :-
        nl,write('constructing specialised program---'),sp_ttyflush,
        procedures(R,R,Cls1,Pathtable,Pfg),
        nl,nl,write('cleaning up---'),sp_ttyflush,
        cleanup(G,Cls1,Cls).

write_to_file(Cls) :-
        nl,write('Enter filename: '),
        read(F),
        sp_filewrite(F,Cls).

/* Rename each atom in the environment */

renames([],R) :-
	list1(R).
renames([element(_,As)|Cs],R) :-
	most_general(As,C),
	renamepred(C,R),
	renames(Cs,R).
	
renamepred(P,S) :-
	renamepred(P,S,0).

renamepred(A,S,M) :-
	var(S),
	!,
	N is M + 1,
	A =.. [P|_],
	name(P,Plist),
	name(N,Nlist),
	my_append(Plist,[95|Nlist],PNlist), /* 95 = underscore */
	name(Pn,PNlist),
	S = [name(A,N,Pn)|_],
	!.
renamepred(P,[name(P,_,_)|_],_) :-
	!.
renamepred(P,[name(Q,M,_)|S],_) :-
	P =.. [A|_],
	Q =.. [A|_],
	!,
	renamepred(P,S,M).
renamepred(P,[_|S],K) :-
	renamepred(P,S,K).

/*  Generate procedures for the calls  */

procedures([],_,[],_,_) :-
	!.
procedures([name(A,N,Pn)|Rs],R,[Cls|Prs],Env,Pfg) :-
        procedure(A,Env,R,N,Pn,Cls,Pfg),
	!,
        procedures(Rs,R,Prs,Env,Pfg).

procedure(A,Env,Rs,0,_,Cls,Pfg) :- /* no renaming of initial goals */
	!,
	nl,nl,write(A),
        melt(A,Ma),
        bagof1((Ma :- R), resultant(Ma,R), Cls1),
	rename_bodies(Cls1,Env,Pfg,Rs,Cls).
procedure(A,Env,Rs,_,Pn,Cls,Pfg) :-
	nl,write(A),
        melt(A,Ma),
        vars(Ma,Vs),
	%nonvarargs(Ma,Vs),
	%Ma =.. [_|Xs],
	%level1_args(Xs,Ws,[]),
	%nonvarargs1(Ws,Vs),
	H =.. [Pn|Vs],
        bagof1((H :- R), Ma^resultant(Ma,R), Cls1),
	rename_bodies(Cls1,Env,Pfg,Rs,Cls).

nonvarargs(M,Vs) :-
	M =.. [_|Xs],
	nonvarargs1(Xs,Vs).

nonvarargs1([],[]).
nonvarargs1([X|Xs],Vs) :-
	ground(X),
	!,
	nonvarargs1(Xs,Vs).
nonvarargs1([X|Xs],[X|Vs]) :-
	nonvarargs1(Xs,Vs).

level1_args([],L,L).
level1_args([X|Xs],L,L2) :-
	nonvar(X),
	!,
	X =.. [_|Ys],
	d_list(Ys,L,L1),
	level1_args(Xs,L1,L2).
level1_args([X|Xs],[X|L],L1) :-
	level1_args(Xs,L,L1).

d_list([],L,L).
d_list([X|Xs],[X|L],L1) :-
	d_list(Xs,L,L1).

rename_bodies([],_,_,_,[]) :-
	!.
rename_bodies([(H :- R1)|Cls1],Env,Pfg,Rs,[(H :- R2)|Cls]) :-
        make_body(R1,Env,Pfg,Rs,R2),
	!,
	rename_bodies(Cls1,Env,Pfg,Rs,Cls).
rename_bodies([_|Cls1],Env,Pfg,Rs,Cls) :-  	/* couldn't make body */
	rename_bodies(Cls1,Env,Pfg,Rs,Cls).	/* because it fails   */

make_body(B,_,_,_,B) :- 
	var(B),
	!.
make_body((B,Bs),Env,Pfg,Rs,(A,As)) :- 
	!,
        make_body(B,Env,Pfg,Rs,A),
        make_body(Bs,Env,Pfg,Rs,As).
make_body(B,Env,Pfg,Rs,B1) :- 
	prolog_meta(B,B1,P,P1),
	!,
	make_body(P,Env,Pfg,Rs,P1).
make_body(B,Env,Pfg,Rs,A) :-
        buildNewGoal(B,Env,Pfg,Rs,A).

buildNewGoal(G,_,_,_,G) :-
        sp_builtin(G),
        !.
buildNewGoal(G,_,_,_,G) :-
        imported(G),
        !.
buildNewGoal(B,Env,Pfg,Rs,A) :-
        copyterm(B,Cb),
	canonical(Cb),
        rep(Cb,R,Env,Pfg),
        renameGoal(R,B,A,Rs),
        !.
%buildNewGoal(B,_,_,_,fail) :-
%	nl,
%	write(B),
%	write('cannot be renamed.......'),
%	nl,nl.

renameGoal(_,A,A,_) :-
	sp_builtin(A),
	!.
renameGoal(_,A,A,_) :-
	imported(A),
	!.
renameGoal(R,A,A,Rs) :-
        memb2(name(R,0,_),Rs),
	!.
renameGoal(R,B,A,Rs) :-
        memb2(name(R,_,Pn),Rs),
	melt(R,Mr),
	vars(Mr,Vs),
	%nonvarargs(Mr,Vs),
	%Mr =.. [_|Xs],
	%level1_args(Xs,Ws,[]),
	%nonvarargs1(Ws,Vs),
	A =.. [Pn|Vs],
	B = Mr,
	!.
 
most_general([A|As],A) :-
	var(As),
	!.
most_general([_|As],A) :-
	most_general(As,A).

rep(A,B,Pathtable,Pfg) :-
	get_path(A,Pathtable,P,_,_),
	get_reps(P,Pfg,Bs),
	most_general(Bs,B).

get_reps(P,[element(P1,Bs)|_],Bs) :-
	P == P1,
	!.
get_reps(P,[_|Es],Bs) :-
	get_reps(P,Es,Bs).

sp_filewrite(tty,Cls) :-
        !,
	nl,write('Specialised Program:'),nl,nl,
        displayqProcs(Cls),
	nl,nl.
sp_filewrite(F,Cls) :-
        tell(F),
        displayqProcs(Cls),
        told.

/* write the program to current output file  */

displayqProcs([P|Ps]) :-
        displayqClauses(P),
        displayqProcs(Ps).
displayqProcs([]).

